#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from functools import wraps
from typing import Callable
from typing import TypeVar

import numpy as np


# ========================================================================= #
# Mutate Helper                                                             #
# ========================================================================= #


F = TypeVar('F')
T = TypeVar('T')
MutateFnHint = Callable[[T], T]


def check_mutation(fn: F) -> F:
    @wraps(fn)
    def wrapper(value: T, *args, **kwargs):
        mutated = fn(value, *args, **kwargs)
        assert mutated is not value, f'Mutate function: {fn} should return a new value'
        return mutated
    return wrapper


# ========================================================================= #
# Mutate                                                                    #
# ========================================================================= #


@check_mutation
def mutate_flip_bits(a: np.ndarray, p: float = 0.05) -> np.ndarray:
    return a ^ (np.random.random(a.shape) < p)


@check_mutation
def mutate_flip_bit_groups(a: np.ndarray, p: float = 0.05) -> np.ndarray:
    if np.random.random() < 0.5:
        # flip set bits
        return a ^ ((np.random.random(a.shape) < p) & a)
    else:
        # flip unset bits
        return a ^ ((np.random.random(a.shape) < p) & ~a)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from typing import Any
from typing import Dict
from typing import Generic
from typing import List
from typing import Sequence
from typing import TypeVar

from ruck._history import StatsGroup
from ruck._member import Population
from ruck.util._args import HParamsMixin


# ========================================================================= #
# Module                                                                    #
# ========================================================================= #


T = TypeVar('T')


class EaModule(Generic[T], HParamsMixin):

    # OVERRIDABLE DEFAULTS

    def get_stats_groups(self) -> Dict[str, StatsGroup[T, Any]]:
        # additional stats to be recorded
        return {}

    def get_progress_stats(self) -> Sequence[str]:
        # which stats are included in the progress bar
        return ('evals', 'fit:max',)

    # REQUIRED

    def gen_starting_values(self) -> List[T]:
        raise NotImplementedError

    def generate_offspring(self, population: Population[T]) -> Population[T]:
        raise NotImplementedError

    def select_population(self, population: Population[T], offspring: Population[T]) -> Population[T]:
        raise NotImplementedError

    def evaluate_values(self, values: List[T]) -> List[float]:
        raise NotImplementedError


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from argparse import Namespace
from typing import Optional
from typing import Sequence


# ========================================================================= #
# Hyper Parameters                                                          #
# ========================================================================= #


class HParamsMixin(object):

    __hparams = None

    def save_hyperparameters(self, ignore: Optional[Sequence[str]] = None, include: Optional[Sequence[str]] = None):
        import inspect
        import warnings
        # get ignored values
        ignored = set() if (ignore is None) else set(ignore)
        included = set() if (include is None) else set(include)
        assert all(str.isidentifier(k) for k in ignored)
        assert all(str.isidentifier(k) for k in included)
        # get function params & signature
        locals = inspect.currentframe().f_back.f_locals
        params = inspect.signature(self.__class__.__init__)
        # get values
        (self_param, *params) = params.parameters.items()
        # check that self is correct & skip it
        assert self_param[0] == 'self'
        assert locals[self_param[0]] is self
        # get other values
        values = {}
        for k, v in params:
            if k in ignored: continue
            if v.kind == v.VAR_KEYWORD: warnings.warn('variable keywords argument saved, consider converting to explicit arguments.')
            if v.kind == v.VAR_POSITIONAL: warnings.warn('variable positional argument saved, consider converting to explicit named arguments.')
            values[k] = locals[k]
        # get extra values
        for k in included:
            assert k != 'self'
            assert k not in values, 'k has already been included'
            values[k] = locals[k]
        # done!
        self.__hparams = Namespace(**values)

    @property
    def hparams(self):
        return self.__hparams


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import itertools
import logging
from typing import Generic
from typing import Iterator
from typing import Tuple
from typing import TypeVar

import numpy as np
from tqdm import tqdm

from ruck._history import HallOfFame
from ruck._history import Logbook
from ruck._history import StatsGroup
from ruck._member import Member
from ruck._member import Population
from ruck._module import EaModule


log = logging.getLogger(__name__)


T = TypeVar('T')


# ========================================================================= #
# Utils Trainer                                                             #
# ========================================================================= #


def _check_population(population: Population[T], required_size: int) -> Population[T]:
    assert len(population) > 0, 'population must not be empty'
    assert len(population) == required_size, 'population size is invalid'
    assert all(isinstance(member, Member) for member in population), 'items in population are not members'
    return population


# ========================================================================= #
# Evaluate Invalid                                                          #
# ========================================================================= #


def _evaluate_unevaluated(module: EaModule[T], members: Population[T]) -> int:
    # get unevaluated members
    unevaluated = [m for m in members if not m.is_evaluated]
    # get fitness values
    fitnesses = list(module.evaluate_values([m.value for m in unevaluated]))
    # save fitness values
    assert len(unevaluated) == len(fitnesses)
    for m, f in zip(unevaluated, fitnesses):
        m.fitness = f
    # return the count
    return len(unevaluated)


# ========================================================================= #
# Functional Trainer                                                        #
# ========================================================================= #


def yield_population_steps(module: EaModule[T]) -> Iterator[Tuple[int, Population[T], Population[T], int]]:
    # 1. create population
    population = [Member(m) for m in module.gen_starting_values()]
    population_size = len(population)
    population = _check_population(population, required_size=population_size)

    # 2. evaluate population
    evals = _evaluate_unevaluated(module, population)

    # yield initial population
    yield 0, population, population, evals

    # training loop
    for i in itertools.count(1):
        # 1. generate offspring
        offspring = module.generate_offspring(population)
        # 2. evaluate
        evals = _evaluate_unevaluated(module, offspring)
        # 3. select
        population = module.select_population(population, offspring)
        population = _check_population(population, required_size=population_size)

        # yield steps
        yield i, population, offspring, evals


# ========================================================================= #
# Class Trainer                                                             #
# ========================================================================= #


class Trainer(Generic[T]):

    def __init__(
        self,
        generations: int = 100,
        progress: bool = True,
        history_n_best: int = 5,
        offspring_generator=yield_population_steps,
    ):
        self._generations = generations
        self._progress = progress
        self._history_n_best = history_n_best
        self._offspring_generator = offspring_generator
        assert self._history_n_best > 0

    def fit(self, module: EaModule[T]) -> Tuple[Population[T], Logbook[T], HallOfFame[T]]:
        assert isinstance(module, EaModule)
        # history trackers
        logbook, halloffame = self._create_default_trackers(module)
        # progress bar and training loop
        with tqdm(total=self._generations, desc='generation', disable=not self._progress, ncols=120) as p:
            for gen, population, offspring, evals in itertools.islice(self._offspring_generator(module), self._generations):
                # update statistics with new population
                halloffame.update(offspring)
                stats = logbook.record(population, gen=gen, evals=evals)
                # update progress bar
                p.update()
                p.set_postfix({k: stats[k] for k in module.get_progress_stats()})
        # done
        return population, logbook, halloffame.freeze()

    def _create_default_trackers(self, module: EaModule[T]) -> Tuple[Logbook[T], HallOfFame[T]]:
        halloffame = HallOfFame(
            n_best=self._history_n_best,
            maximize=True,
        )
        logbook = Logbook(
            'gen', 'evals',
            fit=StatsGroup(lambda pop: [m.fitness for m in pop], min=np.min, max=np.max, mean=np.mean),
            **module.get_stats_groups()
        )
        return logbook, halloffame


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import random
from functools import wraps
from typing import Callable
from typing import TypeVar

from ruck._member import Population


# ========================================================================= #
# Select Helper                                                             #
# ========================================================================= #


F = TypeVar('F')
T = TypeVar('T')
SelectFnHint = Callable[[Population[T], int], Population[T]]


def check_selection(fn: F) -> F:
    @wraps(fn)
    def wrapper(population: Population[T], num: int, *args, **kwargs) -> Population[T]:
        selected = fn(population, num, *args, **kwargs)
        assert selected is not population, f'Select function: {fn} should return a new list'
        assert len(selected) == num, f'Select function: {fn} returned an incorrect number of elements, got: {len(selected)}, should be: {num}'
        return selected
    return wrapper


# ========================================================================= #
# Select                                                                    #
# ========================================================================= #


@check_selection
def select_best(population: Population[T], num: int) -> Population[T]:
    return sorted(population, key=lambda m: m.fitness, reverse=True)[:num]


@check_selection
def select_worst(population: Population[T], num: int) -> Population[T]:
    return sorted(population, key=lambda m: m.fitness, reverse=False)[:num]


@check_selection
def select_random(population: Population[T], num: int) -> Population[T]:
    return random.sample(population, k=num)


@check_selection
def select_tournament(population: Population[T], num: int, k: int = 3) -> Population[T]:
    key = lambda m: m.fitness
    return [
        max(random.sample(population, k=k), key=key)
        for _ in range(num)
    ]


# ========================================================================= #
# Selection                                                                 #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from functools import wraps
from typing import Callable
from typing import Tuple
from typing import TypeVar

import numpy as np


# ========================================================================= #
# Mate Helper                                                               #
# ========================================================================= #


F = TypeVar('F')
T = TypeVar('T')
MateFnHint = Callable[[T, T], Tuple[T, T]]


def check_mating(fn: F) -> F:
    @wraps(fn)
    def wrapper(value_a: T, value_b: T, *args, **kwargs) -> Tuple[T, T]:
        mated_a, mated_b = fn(value_a, value_b, *args, **kwargs)
        assert mated_a is not value_a, f'Mate function: {fn} should return new values'
        assert mated_a is not value_b, f'Mate function: {fn} should return new values'
        assert mated_b is not value_a, f'Mate function: {fn} should return new values'
        assert mated_b is not value_b, f'Mate function: {fn} should return new values'
        return mated_a, mated_b
    return wrapper


# ========================================================================= #
# Mate                                                                      #
# ========================================================================= #


@check_mating
def mate_crossover_1d(a: np.ndarray, b: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    assert a.ndim == 1
    assert a.shape == b.shape
    i, j = np.random.randint(0, len(a), size=2)
    i, j = min(i, j), max(i, j)
    new_a = np.concatenate([a[:i], b[i:j], a[j:]], axis=0)
    new_b = np.concatenate([b[:i], a[i:j], b[j:]], axis=0)
    return new_a, new_b


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import functools
from typing import Any
from typing import List
from typing import Sequence

import ray
from ray.remote_function import RemoteFunction


# ========================================================================= #
# ray                                                                       #
# ========================================================================= #


@functools.lru_cache(maxsize=16)
def _to_remote_ray_fn(fn):
    if not isinstance(fn, RemoteFunction):
        fn = ray.remote(fn)
    return fn


def ray_map(ray_fn, items: Sequence[Any]) -> List[Any]:
    """
    A more convenient alternative to `ray.util.multiprocessing.Pool`s `map` function!
    Using a similar API to python `map`, except returning a list of mapped values
    instead of an iterable.

    The advantage of this functions it that we automatically wrap passed functions to
    ray.remote functions, also enabling automatic getting of ObjectRef values.
    """
    # make sure the function is a remote function
    ray_fn = _to_remote_ray_fn(ray_fn)
    # pass each item to ray and wait for the result
    return ray.get(list(map(ray_fn.remote, items)))


# ========================================================================= #
# ray - object store                                                        #
# ========================================================================= #


def ray_refs_wrapper(fn = None, get: bool = True, put: bool = True, iter_results: bool = False):
    """
    Wrap a function so that we automatically ray.get
    all the arguments and ray.put the result.

    iter_results=True instead treats the result as an
    iterable and applies ray.put to each result item

    for example:
    >>> def mate(a, b):
    >>>    a, b = ray.get(a), ray.get(b)
    >>>    a, b = R.mate_crossover_1d(a, b)
    >>>    return ray.put(a), ray.put(b)

    becomes:
    >>> @ray_refs_wrapper(iter_results=True)
    >>> def mate(a, b):
    >>>     return R.mate_crossover_1d(a, b)
    """

    def wrapper(fn):
        @functools.wraps(fn)
        def inner(*args):
            # get values from object store
            if get:
                args = (ray.get(v) for v in args)
            # call function
            result = fn(*args)
            # store values in the object store
            if put:
                if iter_results:
                    result = tuple(ray.put(v) for v in result)
                else:
                    result = ray.put(result)
            # done!
            return result
        return inner

    # handle correct case
    if fn is None:
        return wrapper
    else:
        return wrapper(fn)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
import re
from typing import Generic
from typing import List
from typing import Optional
from typing import TypeVar

import numpy as np


# ========================================================================= #
# Members                                                                   #
# ========================================================================= #


class MemberIsNotEvaluatedError(Exception):
    pass


class MemberAlreadyEvaluatedError(Exception):
    pass


T = TypeVar('T')


_RE_WHITESPACE = re.compile(r'\s\s+')


class Member(Generic[T]):

    def __init__(self, value: T, fitness: float = None):
        self._value = value
        self._fitness = None
        # set fitness
        if fitness is not None:
            self.fitness = fitness

    @property
    def value(self) -> T:
        return self._value

    @property
    def fitness_unsafe(self) -> Optional[float]:
        return self._fitness

    @property
    def fitness(self) -> float:
        if not self.is_evaluated:
            raise MemberIsNotEvaluatedError('The member has not been evaluated, the fitness has not yet been set.')
        return self._fitness

    @fitness.setter
    def fitness(self, fitness: float):
        if self.is_evaluated:
            raise MemberAlreadyEvaluatedError('The member has already been evaluated, the fitness can only ever be set once. Create a new member instead!')
        if np.isnan(fitness):
            raise ValueError('fitness values cannot be NaN, this is an error!')
        self._fitness = float(fitness)

    def set_fitness(self, fitness: float) -> 'Member[T]':
        self.fitness = fitness
        return self

    @property
    def is_evaluated(self) -> bool:
        return (self._fitness is not None)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        value_str = _RE_WHITESPACE.sub(' ', repr(self.value))
        # cut short
        if len(value_str) > 33:
            value_str = f'{value_str[:14].rstrip(" ")} ... {value_str[-14:].lstrip(" ")}'
        # get fitness
        fitness_str = f', {self.fitness}' if self.is_evaluated else ''
        # combine
        return f'{self.__class__.__name__}({value_str}{fitness_str})'


# ========================================================================= #
# Population                                                                #
# ========================================================================= #


Population = List[Member[T]]


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import dataclasses
import heapq
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generic
from typing import List
from typing import TypeVar

from ruck._member import Population


T = TypeVar('T')
V = TypeVar('V')


# ========================================================================= #
# Type Hints                                                                #
# ========================================================================= #


ValueFnHint     = Callable[[T], V]
StatFnHint      = Callable[[V], Any]


# ========================================================================= #
# Logbook                                                                   #
# ========================================================================= #


class StatsGroup(Generic[T, V]):

    def __init__(self, value_fn: ValueFnHint[T, V] = None, **stats_fns: StatFnHint[V]):
        assert all(str.isidentifier(key) for key in stats_fns.keys())
        assert stats_fns
        self._value_fn = value_fn
        self._stats_fns = stats_fns

    @property
    def keys(self) -> List[str]:
        return list(self._stats_fns.keys())

    def compute(self, value: T) -> Dict[str, Any]:
        if self._value_fn is not None:
            value = self._value_fn(value)
        return {
            key: stat_fn(value)
            for key, stat_fn in self._stats_fns.items()
        }


class Logbook(Generic[T]):

    def __init__(self, *external_keys: str, **stats_groups: StatsGroup[T, Any]):
        self._all_ordered_keys = []
        self._external_keys = []
        self._stats_groups = {}
        self._history = []
        # register values
        for k in external_keys:
            self.register_external_stat(k)
        for k, v in stats_groups.items():
            self.register_stats_group(k, v)

    def _assert_key_valid(self, name: str):
        if not str.isidentifier(name):
            raise ValueError(f'stat name is not a valid identifier: {repr(name)}')
        return name

    def _assert_key_available(self, name: str):
        if name in self._external_keys:
            raise ValueError(f'external stat already named: {repr(name)}')
        if name in self._stats_groups:
            raise ValueError(f'stat group already named: {repr(name)}')
        return name

    def register_external_stat(self, name: str):
        self._assert_key_available(self._assert_key_available(name))
        # add stat
        self._external_keys.append(name)
        self._all_ordered_keys.append(name)
        return self

    def register_stats_group(self, name: str, stats_group: StatsGroup[T, Any]):
        self._assert_key_available(self._assert_key_available(name))
        assert isinstance(stats_group, StatsGroup)
        assert stats_group not in self._stats_groups.values()
        # add stat group
        self._stats_groups[name] = stats_group
        self._all_ordered_keys.extend(f'{name}:{key}' for key in stats_group.keys)
        return self

    def record(self, population: Population[T], **external_values):
        # extra stats
        if set(external_values.keys()) != set(self._external_keys):
            raise KeyError(f'required external_values: {sorted(self._external_keys)}, got: {sorted(external_values.keys())}')
        # external values
        stats = dict(external_values)
        # generate stats
        for name, stat_group in self._stats_groups.items():
            for key, value in stat_group.compute(population).items():
                stats[f'{name}:{key}'] = value
        # order stats
        assert set(stats.keys()) == set(self._all_ordered_keys)
        record = {k: stats[k] for k in self._all_ordered_keys}
        # record and return stats
        self._history.append(record)
        return dict(record)

    @property
    def history(self) -> List[Dict[str, Any]]:
        return list(self._history)

    def __getitem__(self, idx: int):
        assert isinstance(idx, int)
        return dict(self._history[idx])

    def __len__(self):
        return len(self._history)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]


# ========================================================================= #
# HallOfFame                                                                #
# ========================================================================= #


@dataclasses.dataclass(order=True)
class HallOfFameItem:
    fitness: float
    member: Any = dataclasses.field(compare=False)


class HallOfFameFrozenError(Exception):
    pass


class HallOfFameNotFrozenError(Exception):
    pass


class HallOfFame(Generic[T]):

    def __init__(self, n_best: int = 5, maximize: bool = True):
        self._maximize = maximize
        assert maximize
        self._n_best = n_best
        # update values
        self._heap = []  # element 0 is always the smallest
        self._scores = {}
        # frozen values
        self._frozen = False
        self._frozen_members = None
        self._frozen_values = None
        self._frozen_scores = None

    def update(self, population: Population[T]):
        if self.is_frozen:
            raise HallOfFameFrozenError('The hall of fame has been frozen, no more members can be added!')
        # get potential best in population
        best = sorted(population, key=lambda m: m.fitness, reverse=True)[:self._n_best]
        # add the best
        for member in best:
            # try add to hall of fame
            item = HallOfFameItem(fitness=member.fitness, member=member)
            # skip if we already have the same score ...
            # TODO: this should not ignore members with the same scores, this is hacky
            if item.fitness in self._scores:
                continue
            # checks
            self._scores[item.fitness] = item
            if len(self._heap) < self._n_best:
                heapq.heappush(self._heap, item)
            else:
                removed = heapq.heappushpop(self._heap, item)
                del self._scores[removed.fitness]

    def freeze(self) -> 'HallOfFame':
        if self.is_frozen:
            raise HallOfFameFrozenError('The hall of fame has already been frozen, cannot freeze again!')
        # freeze
        self._frozen = True
        self._frozen_members = [m.member for m in sorted(self._heap, reverse=True)]  # 0 is best, -1 is worst
        # reset values
        self._scores = None
        self._heap = None
        return self

    @property
    def is_frozen(self) -> bool:
        return self._frozen

    @property
    def members(self) -> Population[T]:
        return list(self._frozen_members)

    def __getitem__(self, idx: int):
        if not self.is_frozen:
            raise HallOfFameNotFrozenError('The hall of fame has not yet been frozen by a completed training run, cannot access members!')
        assert isinstance(idx, int)
        return self._frozen_members[idx]

    def __len__(self):
        if not self.is_frozen:
            raise HallOfFameNotFrozenError('The hall of fame has not yet been frozen by a completed training run, cannot access length!')
        return len(self._frozen_members)

    def __iter__(self):
        if not self.is_frozen:
            raise HallOfFameNotFrozenError('The hall of fame has not yet been frozen by a completed training run, cannot access members!')
        for i in range(len(self)):
            yield self[i]



# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import setuptools


# ========================================================================= #
# HELPER                                                                    #
# ========================================================================= #


with open("README.md", "r", encoding="utf-8") as file:
    long_description = file.read()

with open('requirements.txt', 'r') as f:
    install_requires = (req[0] for req in map(lambda x: x.split('#'), f.readlines()))
    install_requires = [req for req in map(str.strip, install_requires) if req]


# ========================================================================= #
# SETUP                                                                     #
# ========================================================================= #


setuptools.setup(
    name="ruck",
    author="Nathan Juraj Michlo",
    author_email="NathanJMichlo@gmail.com",

    version="0.0.1.dev1",
    python_requires=">=3.8",
    packages=setuptools.find_packages(),

    install_requires=install_requires,

    url="https://github.com/nmichlo/ruck",
    description="Performant evolutionary algorithms for Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",

    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Intended Audience :: Science/Research",
    ],
)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from typing import Optional
from typing import Tuple
from typing import TypeVar

from ruck._member import Member
from ruck._member import Population
from ruck.functional import SelectFnHint
from ruck.functional._mate import MateFnHint
from ruck.functional._mutate import MutateFnHint
from ruck.util._iter import replaced_random_taken_pairs
from ruck.util._iter import replaced_random_taken_elems


import random
import numpy as np


# ========================================================================= #
# Helper                                                                    #
# ========================================================================= #


T = TypeVar('T')


# ========================================================================= #
# Crossover & Mutate Helpers                                                #
# ========================================================================= #


def _mate_wrap_unwrap_values(mate_fn: MateFnHint[T]):
    def wrapper(ma: Member[T], mb: Member[T]) -> Tuple[Member[T], Member[T]]:
        va, vb = mate_fn(ma.value, mb.value)
        return Member(va), Member(vb)
    return wrapper


def _mutate_wrap_unwrap_values(mutate_fn: MutateFnHint[T]):
    def wrapper(m: Member[T]) -> Member[T]:
        v = mutate_fn(m.value)
        return Member(v)
    return wrapper


# ========================================================================= #
# Function Wrappers                                                         #
# ========================================================================= #


def apply_mate(
    population: Population[T],
    mate_fn: MateFnHint[T],
    p: float = 0.5,
    keep_order: bool = True,
    map_fn=map,
) -> Population[T]:
    # randomize order so we have randomized pairs
    if keep_order:
        indices = np.arange(len(population))
        np.random.shuffle(indices)
        offspring = [population[i] for i in indices]
    else:
        offspring = list(population)
        np.random.shuffle(offspring)
    # apply mating to population
    offspring = replaced_random_taken_pairs(
        fn=_mate_wrap_unwrap_values(mate_fn),
        items=offspring,
        p=p,
        map_fn=map_fn,
    )
    # undo random order
    if keep_order:
        offspring = [offspring[i] for i in np.argsort(indices)]
    # done!
    assert len(offspring) == len(population)
    return offspring


def apply_mutate(
    population: Population,
    mutate_fn: MutateFnHint,
    p: float = 0.5,
    map_fn=map,
) -> Population:
    # apply mutations to population
    offspring = replaced_random_taken_elems(
        fn=_mutate_wrap_unwrap_values(mutate_fn),
        items=population,
        p=p,
        map_fn=map_fn,
    )
    # done!
    assert len(offspring) == len(population)
    return offspring


def apply_mate_and_mutate(
    population: Population[T],
    mate_fn: MateFnHint[T],
    mutate_fn: MutateFnHint[T],
    p_mate: float = 0.5,
    p_mutate: float = 0.5,
    map_fn=map,
) -> Population[T]:
    """
    Apply crossover AND mutation

    NOTE:
    - Modified individuals need their fitness re-evaluated
    - Mate & Mutate should always return copies of the received values.

    ** Should be equivalent to varAnd from DEAP **
    """
    offspring = apply_mate(population, mate_fn, p=p_mate, keep_order=True, map_fn=map_fn)
    offspring = apply_mutate(offspring, mutate_fn, p=p_mutate, map_fn=map_fn)
    return offspring


def _get_generate_member_fn(
    mate_fn: MateFnHint[T],
    mutate_fn: MutateFnHint[T],
    p_mate: float = 0.5,
    p_mutate: float = 0.5,
):
    def _generate_member(a_b_r: Tuple[Member[T], Optional[Member[T]], float]) -> Member[T]:
        ma, mb, r = a_b_r
        if   r < p_mate:            return Member(mate_fn(ma.value, mb.value)[0])  # Apply crossover | only take first item | mb is only defined for this case
        elif r < p_mate + p_mutate: return Member(mutate_fn(ma.value))             # Apply mutation
        else:                       return ma                                      # Apply reproduction
    return _generate_member


def apply_mate_or_mutate_or_reproduce(
    population: Population[T],
    num_offspring: int,  # lambda_
    mate_fn: MateFnHint[T],
    mutate_fn: MutateFnHint[T],
    p_mate: float = 0.5,
    p_mutate: float = 0.5,
    map_fn=map,
) -> Population[T]:
    """
    Apply crossover OR mutation OR reproduction

    NOTE:
    - Modified individuals need their fitness re-evaluated
    - Mate & Mutate should always return copies of the received values.

    ** Should be equivalent to varOr from DEAP, but significantly faster for larger populations **
    """
    assert (p_mate + p_mutate) <= 1.0, 'The sum of the crossover and mutation probabilities must be smaller or equal to 1.0.'

    # choose which action should be taken for each element
    probabilities  = np.random.random(num_offspring)
    # select offspring
    choices_a = [random.choice(population)                           for p in probabilities]
    choices_b = [random.choice(population) if (p < p_mate) else None for p in probabilities]  # these are only needed for crossover, when (p < p_mate)
    # get function to generate offspring
    # - we create the function so that we don't accidentally pickle anything else
    fn = _get_generate_member_fn(mate_fn=mate_fn, mutate_fn=mutate_fn, p_mate=p_mate, p_mutate=p_mutate)
    # generate offspring
    # - TODO: this is actually not optimal! we should only pass mate and
    #         mutate operations to the map function, we could distribute
    #         work unevenly between processes if map_fn is replaced
    offspring = list(map_fn(fn, zip(choices_a, choices_b, probabilities)))
    # done!
    assert len(offspring) == num_offspring
    return offspring


# ========================================================================= #
# Gen & Select                                                              #
# ========================================================================= #


def factory_simple_ea(
    mate_fn: MateFnHint[T],
    mutate_fn: MutateFnHint[T],
    select_fn: SelectFnHint[T],
    p_mate: float = 0.5,
    p_mutate: float = 0.5,
    map_fn=map,
):
    def generate(population):
        return apply_mate_and_mutate(population=select_fn(population, len(population)), p_mate=p_mate, mate_fn=mate_fn, p_mutate=p_mutate, mutate_fn=mutate_fn, map_fn=map_fn)

    def select(population, offspring):
        return offspring

    return generate, select


def factory_mu_plus_lambda(
    mate_fn: MateFnHint[T],
    mutate_fn: MutateFnHint[T],
    select_fn: SelectFnHint[T],
    offspring_num: int,   # lambda
    p_mate: float = 0.5,
    p_mutate: float = 0.5,
    map_fn=map,
):
    def generate(population):
        num = len(population) if (offspring_num is None) else offspring_num
        return apply_mate_or_mutate_or_reproduce(population, num, mate_fn=mate_fn, mutate_fn=mutate_fn, p_mate=p_mate, p_mutate=p_mutate, map_fn=map_fn)

    def select(population: Population[T], offspring: Population[T]):
        return select_fn(population + offspring, len(population))

    return generate, select


def factory_mu_comma_lambda(
    mate_fn: MateFnHint[T],
    mutate_fn: MutateFnHint[T],
    select_fn: SelectFnHint[T],
    offspring_num: Optional[int] = None, # lambda
    p_mate: float = 0.5,
    p_mutate: float = 0.5,
    map_fn=map,
):
    def generate(population):
        num = len(population) if (offspring_num is None) else offspring_num
        return apply_mate_or_mutate_or_reproduce(population, num, mate_fn=mate_fn, mutate_fn=mutate_fn, p_mate=p_mate, p_mutate=p_mutate, map_fn=map_fn)

    def select(population, offspring):
        assert len(offspring) >= len(population), f'invalid arguments, the number of offspring: {len(offspring)} (lambda) must be greater than or equal to the size of the population: {len(population)} (mu)'
        return select_fn(offspring, len(population))

    return generate, select


# # ========================================================================= #
# # END                                                                       #
# # ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~


import itertools
import random
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Sequence
from typing import Tuple
from typing import TypeVar

import numpy as np


# ========================================================================= #
# Helper                                                                    #
# ========================================================================= #


T = TypeVar('T')


# ========================================================================= #
# iter                                                                      #
# ========================================================================= #


# NOTE:
#   Iterable: objects that return Iterators when passed to `iter()`
#   Iterator: return the next item when used with `next()`
#             every Iterator is ALSO an Iterable


def ipairs(items: Iterable[T]) -> Iterator[Tuple[T, T]]:
    itr_a, itr_b = itertools.tee(items)
    itr_a = itertools.islice(itr_a, 0, None, 2)
    itr_b = itertools.islice(itr_b, 1, None, 2)
    return zip(itr_a, itr_b)


# ========================================================================= #
# lists                                                                     #
# ========================================================================= #


def chained(list_of_lists: Iterable[Iterable[T]]) -> List[T]:
    return list(itertools.chain(*list_of_lists))


def splits(items: Sequence[Any], num_chunks: int, keep_empty: bool = False) -> List[List[Any]]:
    # np.array_split will return empty elements if required
    if not keep_empty:
        num_chunks = min(num_chunks, len(items))
    # we return a lists of lists, not a list of
    # tuples so that it is compatible with ray.get
    return [list(items) for items in np.array_split(items, num_chunks)]


# ========================================================================= #
# random -- used for ruck.functional._algorithm                             #
# ========================================================================= #


def replaced_random_taken_pairs(fn: Callable[[T, T], Tuple[T, T]], items: Iterable[T], p: float, map_fn=map) -> List[T]:
    # shallow copy because we want to update elements in this list
    # - we need to take care to handle the special case where the length
    #   of items is odd, thus we cannot just call random_map with modified
    #   args using pairs and chaining the output
    items = list(items)
    # select random items
    idxs, vals = [], []
    for i, pair in enumerate(zip(items[0::2], items[1::2])):
        if random.random() < p:
            vals.append(pair)
            idxs.append(i)
    # map selected values
    vals = map_fn(lambda pair: fn(pair[0], pair[1]), vals)
    # update values
    for i, (v0, v1) in zip(idxs, vals):
        items[i*2+0] = v0
        items[i*2+1] = v1
    # done!
    return items


def replaced_random_taken_elems(fn: Callable[[T], T], items: Iterable[T], p: float, map_fn=map) -> List[T]:
    # shallow copy because we want to update elements in this list
    items = list(items)
    # select random items
    idxs, vals = [], []
    for i, v in enumerate(items):
        if random.random() < p:
            vals.append(v)
            idxs.append(i)
    # map selected values
    vals = map_fn(fn, vals)
    # update values
    for i, v in zip(idxs, vals):
        items[i] = v
    # done!
    return items


# ========================================================================= #
# END                                                                       #
# ========================================================================= #
